"""Python junk drawer."""
# pyright:reportGeneralTypeIssues=information
import sys
import abc
import time
import types
import array
import ctypes
import typing
import inspect
import itertools
import functools
import threading
from .common import Callable, Any, Iterable, Mapping, Sequence, MutableSequence, MutableMapping


_T = typing.TypeVar("_T")
_O = typing.TypeVar("_O")
_P = typing.ParamSpec("_P")

_SENTINEL = object()





def is_builtin(o: Any) -> bool:
    """Return `True` if an object is a built-in class or function.

    A 'built-in' object is any object available in the `builtins` module.
    This includes all primitive types (`str`, `list`, etc.) and functions
    such as `print` and `isinstance`.
    """
    try:
        return o.__module__ == type.__module__
    except:
        return False


def is_type_object(o: Any) -> bool:
    """Return `True` if an object is a class/type.

    A class/type object is any object that is, or is an instance of, the `type`
    built-in class.
    """
    return isinstance(o, type)


def is_inst_object(o: Any) -> bool:
    """Return `True` if an object is an instance object.

    An instance object is any object that is not, or is not an instance of, the
    `type` built-in class.
    """
    return not isinstance(o, type)


def is_instance(o: Any, tp: type[Any] | tuple[type[Any], ...], /) -> bool:
    """Return `True` if an object is an instance of a type, or is an instance
    of any of the included types. Returns `False` otherwise, or if the object
    is not an instance.

    Args:
        * o: The class/type or instance object to test.

        * tp: A single class/type, or a tuple of class/type objects.


    There is no functional difference between this function and the `isinstance`
    built-in.
    """
    # unlike `issubclass`, no TypeError is thrown due to the object's
    # classification
    return isinstance(o, tp)


def is_subclass(o: Any, tp: type[Any] | tuple[type[Any], ...], /) -> bool:
    """Return `True` if an object is (or is a subclass of) a type, or is (or is a
    subclass of) any of the included types. Returns `False` otherwise, or if the
    object is not a class.

    Args:
        * o: The class/type object to test.

        * tp: A single class/type, or a tuple of class/type objects.


    The only functional difference between this function and the `isinstance`
    built-in is that the `TypeError` is handled when *o* is not a class.
    """
    try:
        return issubclass(o, tp)
    except TypeError:
        if is_type_object(o):
            raise
        return False


def is_iterable(o: Any):
    """Return `True` if an object has an `.__iter__` method.

    An object is iterable if the result of `iter(object)` returns an iterator.
    All iterators, sequences, and containers are iterable.
    """
    return hasattr(o, "__iter__")


def is_iterator(o: Any):
    """Return `True` if an object is iterable and has a `.__next__` method.

    All iterators are iterable, but not all iterables are iterators. For
    example, `next([0, 1, 2])` results in an error, but `next(iter([0, 1, 2]))`
    does not.
    """
    try:
        o.__next__
        o.__iter__
    except:
        return False
    return True


def is_container(o: Any, *, protocol_only: bool = False):
    """Return `True` if an object is a container.

    Containers are iterables capable of storing various objects. They are
    required to implement the `.__contains__` method.
    """
    return hasattr(o, "__contains__")


def is_sequence(o: Any):
    """Return `True` if an object is a sequence.

    Sequences are subscriptable containers storing ordered values and feature
    position-based lookups -- typically using integers or integer-like keys as
    subscript -- called indexes. They are required to implement the `.__len__` and
    `.__getitem__` methods, where `.__getitem__` differs to the subscript's
    `.__index__` method for lookups.

    Mappings and sequences share the same protocol. A sequence's `.__getitem__`
    method will raise 'IndexError' when the subscript is out-of-range, where
    a mapping's `.__getitem__` method will raise `KeyError`.
    """
    try:
        o[len(o) + 1]
    except IndexError:
        return True
    except:
        pass
    return False


def is_mapping(o: Any):
    """Return `True` if an object is a mapping.

    Mappings are subscriptable containers storing associated key-value pairs
    and support arbitrary key lookups, allowing any hashable object to be used as
    a key. They are required to implement the `.__len__` and `.__getitem__` methods,
    where `.__getitem__` differs to the subscript's `.__hash__` method for lookups.

    Mappings and sequences share the same protocol.
    """
    try:
        o.__len__
        o[_SENTINEL]
    except KeyError:
        return True
    except:
        pass
    return False


def dunders(tp: type[Any], *, excludes: Sequence[str] = (), excl_inherited: bool = False) -> list[str]:
    """Return the names of an object's dunder methods.

    Args:
        * o: The object of inquiry.

        * excludes: Names of members to exclude.

        * excl_inherited: If `True`, only fetch members from the type's
        `__dict__`.
    """
    startsw = str.startswith
    endsw   = str.endswith
    # `.startswith("__")` is *usually* enough, since most that also
    # don't end with "__" are name-mangled. But just in case...
    return sorted(set(
        n
        for n in (tp.__dict__ if excl_inherited else dir(tp))
        if startsw(n, "__") and endsw(n, "__") and n not in excludes
    ))


def flatten(tp: type[_T], *iterables: Iterable[Any]) -> _T:
    """Return a flattened collection object from iterables."""
    return tp(itertools.chain(*iterables))


def flatten_as_set(*iterables: Iterable[Any]) -> set[Any]:
    """Return a flattened set from several iterables."""
    return flatten(set, *iterables)


def flatten_as_list(*iterables: Iterable[Any]) -> list[Any]:
    """Return a flattened list from several iterables."""
    return flatten(list, *iterables)


def flatten_as_tuple(*iterables: Iterable[Any]) -> tuple[Any]:
    """Return a flattened tuple from several iterables."""
    return flatten(tuple, *iterables)


def flatten_as_arr(tp_code: str, *iterables: Iterable[Iterable[_T]]) -> array.array | MutableSequence[_T]:
    """Return a flattened Python array from several iterables."""
    return array.array(tp_code, itertools.chain(*iterables))


def hasattrs(o: Any, *names: str) -> bool:
    """Return True if an object has attribute(s) of all given names."""
    return all(hasattr(o, name) for name in names)


def namespace_path(o: type | types.FunctionType) -> tuple[str, str]:
    assert hasattr(sys.modules[o.__module__], o.__name__)
    return o.__module__, o.__name__


def create_function(name: str, args: Sequence[str], body: Sequence[str], return_type: _T = Any, *, globals: dict | None = None, locals: dict | None = None) -> Callable[..., _T]:
    """Dynamically define and return a new function.

    Args:
        * name: Formal name for the new function.

        * args: A sequence containing argument signatures (as strings) for the new
        function. Each value in *args* should be limited to a single argument
        signature.

        * body: A sequence containing the source code (as strings) that will execute
        when the new function is called. Each value in *body* should be limited to
        one line of executable source code.

        * return_tp: The type/annotation indicating the new function's return type.

        * globals: The global scope that will be available to the new function.

        * locals: The local scope that will be available to the new function.
    """
    locals = locals if locals is not None else {}
    locals["_return"] = return_type

    body    = '\n'.join(f"        {line}" for line in body)
    closure = (
        f"def __create_function__({', '.join(locals)}):\n"
        f"    def {name}({', '.join(args)}) -> _return:\n{body}\n"
        f"    return {name}"
    )

    scope = {}
    exec(closure, globals, scope)

    fn = scope["__create_function__"](**locals)
    fn.__qualname__ = name
    return fn

@typing.final
class _MarkerType(typing.Protocol):
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: Any) -> bool: ...
    def __hash__(self) -> int: ...
    def __delattr__(self, name: str) -> TypeError: ...
    def __setattr__(self, name: str, value: Any) -> TypeError: ...


def create_marker_type(name: str, *, body: Mapping[str, Any] | Sequence[tuple[str, Any]] = ()) -> type[_MarkerType]:
    """Dynamically create and return a new marker type as if it were defined
    in the module this function is called from.

    Markers cannot be instantiated or subclassed, and are read-only by default.

    Args:
        * name: Class name of the marker.

        * body: A mapping to use as the namespace/dict of the marker's
        metaclass.


    This function creates both the marker class as well as the marker's metaclass.
    By default, the metaclass implements the `__repr__`, `__eq__`, `__hash__`,
    `__setattr__`, and `__delattr__` methods. User implementations included in
    *body* will be used instead, if present.

    Note that the metaclass also implements `__new__` and `__call__`. User
    implementations of these methods will be used when creating the marker class
    if available. However, they are deleted from the metaclass' dict before
    returning the marker class.
    """
    body = dict(body)
    if '__module__' in body:
        module = sys.modules[body['__module__']]
    else:
        module = inspect.getmodule(inspect.stack()[1][0])
        assert module.__name__  # type:ignore

    namespace = {
        '__module__': module.__name__,  # type:ignore
        '__repr__': create_function(
            '__repr__',
            ('cls',),
            (f'return "<class \'{name}\'>"',),
            globals=module.__dict__
        ),
        '__eq__': create_function(
            '__eq__',
            ('cls', 'other'),
            (f'return other is cls',),
            globals=module.__dict__
        ),
        '__hash__': create_function(
            '__hash__',
            ('cls',),
            (f'return {hash(object())}',),
            globals=module.__dict__,
        ),
        '__setattr__': create_function(
            '__setattr__',
            ('cls', 'name = None', 'value = None'),
            (f"raise TypeError(\"cannot set attributes on read-only class {name!r}.\")",),
            globals=module.__dict__,
        ),
        '__delattr__': create_function(
            '__delattr__',
            ('cls', 'name = None'),
            (f"raise TypeError(\"cannot delete attributes on read-only class {name!r}.\")",),
            globals=module.__dict__,
        ),
    }
    namespace.update(body)
    cls: Any = type(f'{name.title()}Type', (type,), namespace)(
        name, (), {'__module__': module.__name__}   # type:ignore
    )
    type(cls).__call__ = create_function(
        '__call__',
        ('*args', '**kwargs'),
        (f"raise TypeError(\"{name!r} class cannot be instantiated.\")",),
        globals=module.__dict__,
    )
    type(cls).__new__ = create_function(
        '__new__',
        ('*args', '**kwargs'),
        (f"raise TypeError(\"one or more bases cannot be subclassed.\")",),
        globals=module.__dict__,
    )
    return typing.final(cls)


def frozen_namespace(cls: type[_T]) -> type[_T]:
    contents = tuple(m for m in dir(cls) if not m.startswith('_'))
    return create_marker_type(  # type: ignore
        cls.__qualname__,
        body={
            "__module__": cls.__module__,
            '__annotations__': cls.__annotations__,
            "__iter__": create_function(
                '__iter__',
                ('self',),
                ('return iter({' + ', '.join(f"'{m}': self.{m}" for m in contents) + '}.items())',),
                globals=sys.modules[cls.__module__].__dict__,
            ),
            **{m: getattr(cls, m) for m in contents}
        }
    )




def cache_once(fn: Callable[_P, _T], /) -> Callable[_P, _T]:
    """Similar to `functools.cached_property` but for functions. Returns the
    same result regardless of the arguments passed to the decorated function.
    """
    @functools.wraps(fn)
    def cached_result(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        nonlocal cache
        try:
            return cache
        except NameError:
            cache = fn(*args, **kwargs)
            return cache
    cache: _T
    return cached_result


@functools.lru_cache(128)
def parameters(c: Callable) -> tuple[inspect.Parameter]:
    """Return the parameters of the given callable. The result
    is cached for future calls.

    Args:
        * c: Callable for the query.
    """
    return tuple(inspect.signature(c).parameters.values())










class _property(typing.Generic[_T], abc.ABC):
    __slots__ = ('__wrapped__',)

    def __init__(self, fget: Callable[..., _T]):
        self.__wrapped__ = fget

    def __set_name__(self, cls: type[Any], name: str):
        set_name = getattr(self.__wrapped__, "__set_name__", None)
        if set_name:
            set_name(cls, name)

    @abc.abstractmethod
    def __get__(self, instance: Any, cls: type | None = None) -> _T | typing.Self: ...


class classproperty(_property, typing.Generic[_T]):
    """A descriptor for emulating class-bound virtual attributes. Can
    optionally be used as a decorator.

    The functionality of this descriptor is similar to a `classmethod` and
    `property` descriptor chain in Python 3.9 which, as of 3.11, has been
    deprecated.

    Note that, due to Python's lookup behavior, the mutability of the
    property can be misleading; it is read-only when an assignment is made
    on an instance, but not on the class. This weird behavior is one of the
    reasons for the feature's deprecation.
    """
    __slots__ = ()

    def __get__(self, instance: Any, cls: type[Any]) -> _T:
        return self.__wrapped__(cls)


class staticproperty(_property, typing.Generic[_T]):
    """A descriptor for emulating unbound virtual attributes. Can optionally
    be used as a decorator.

    The functionality of this descriptor is similar to a `staticmethod` and
    `property` descriptor chain in Python 3.9 which, as of 3.11, has been
    deprecated.

    Note that, due to Python's lookup behavior, the mutability of the
    property can be misleading; it is read-only when an assignment is made
    on an instance, but not on the class. This weird behavior is one of the
    reasons for the feature's deprecation.
    """
    __slots__ = ()

    def __get__(self, instance: Any, cls: type[Any]) -> _T:
        return self.__wrapped__()


class simpleproperty(_property, typing.Generic[_T]):
    """A data descriptor for emulating read-only virtual attributes. Can
    optionally be used as a decorator.
        >>> class MyClass:
        ...     foo = simpleproperty()
        ...


    The behavior of `foo = simpleproperty()` within a class body is roughly
    outlined in the below example using a `property` descriptor:
        >>> class MyClass:
        ...     @property
        ...     def foo(self):
        ...         return self._foo
        ...
        ...     @foo.setter
        ...     def foo(self, value):
        ...         if hasattr(self, '_foo'):
        ...             raise AttributeError
        ...         self._foo = value

    When accessed, the descriptor returns the value of a similarly-
    named private member on the instance -- That is, a member whose name
    matches the name the descriptor is assigned to, but prefixed with a
    single underscore. Writes to the "public" member are allowed, but
    only if the "private" member is missing. It is otherwise read-only,
    and any attempt to write to the public member are met with
    `AttributeError`:
        >>> my_instance = MyClass()
        >>> my_instance.foo = 'bar'  # WRITABLE
        >>> my_instance.foo
        'bar'
        >>> my_instance.foo = 'BAR'  # READ-ONLY
        Traceback (most recent call last):
            ...
        AttributeError: 'MyClass' object attribute 'foo' is read-only.

    The default setter implementation may not work for all cases, as it
    is usually desirable to transform the value before setting it. The
    descriptor's constructor accepts a custom setter as the first argument,
    allowing it to be used as a decorator:
        >>> class MyClass:
        ...     @simpleproperty
        ...     def foo(self, value):
        ...         self._foo = str(value).lower()
        ...
        >>> my_instance = MyClass()
        >>> my_instance.foo = 'BAR'
        >>> my_instance.foo
        'bar'

    Unlike the outlined `property` simplification above, the logic that
    determines when to perform the set is actually decoupled from the setter
    itself. The descriptor determines *when* to invoke the setter, and the
    setter determines *how* to set the value. That is, if the descriptor
    determines the public member is read-only (i.e. `my_instance._foo` exists),
    it won't call the setter.

    When accessing the public member, the "getter" used by the descriptor
    first checks the instance dictionary for the private member. If it is not
    found, the result of calling `getattr(instance, '_member')` is returned.

    The "getter" used by the descriptor is static; it cannot be configured nor
    can a new one be set. If your use-case extends beyond the capacity of this
    simple descriptor, it is best to use a `property` instead.

    When setting `simpleproperty` descriptors on an existing class, you must
    manually call the `.__set_name__` method to fully initialize the descriptor
    (Python does this automatically DURING class creation). Otherwise,
    `RuntimeError` may be thrown when attempting to access the target member.

    The descriptor works for slotted classes. However, the class must define,
    inherit, or slot the "private" target member.
    """
    __slots__ = ('name',)

    def __init__(self, fset: Callable[[Any, Any], _T] | None = None):
        self.__wrapped__ = fset
        self.name = ''

    def __set_name__(self, cls: type, name: str):
        super().__set_name__(cls, name)
        name = f"_{name}"
        if '__slots__' in cls.__dict__:
            slots = set(itertools.chain(getattr(b, '__slots__', ()) for b in cls.mro()))
            if name not in slots and getattr(cls, name, _SENTINEL) is _SENTINEL:
                raise ValueError(
                    f"slotted class {cls.__qualname__!r} does not include, inherit, "
                    f"or define non-public member {name!r}."
                )
        self.name = name

    def __get__(self, instance: Any, cls: type | None = None) -> _T:
        try:
            return instance.__dict__[self.name]
        except AttributeError:
            if instance is None:
                return self  # type: ignore
            elif not self.name:
                raise RuntimeError(
                    f'{type(self).__qualname__!r} descriptor not fully initialized '
                    f'-- did you forget to call `__set_name__`?'
                )
            return getattr(instance, self.name)
        except KeyError:
            if self.name:
                raise AttributeError(
                    f'{type(instance).__qualname__!r} object has no attribute '
                    f'{self.name[1:]!r}.'
                ) from None
            raise

    def __set__(self, instance: Any, value: Any) -> None:
        if hasattr(instance, self.name):
            raise AttributeError(
                f'{type(instance).__qualname__!r} object attribute {self.name[1:]!r} is read-only.'
            )
        if self.__wrapped__ is None:
            setattr(instance, self.name, value)
        else:
            self.__wrapped__(instance, value)

    def __copy__(self) -> typing.Self:
        copy = type(self)(self.__wrapped__)
        copy.name = self.name
        return copy

    def setter(self, fset: Callable[[Any, Any], Any] | None) -> typing.Self:
        copy = self.__copy__()
        copy.__wrapped__ = fset
        return copy




# reentrant locks work better with debuggers
# XXX: BEWARE -- this may mask deadlocks in release builds!
Lock = threading.Lock if not sys.gettrace() else threading.RLock

class Locker(typing.Generic[_T]):
    """A mutex-like object that encapsulates a value that is to be managed
    only between select callables.
    """
    __slots__ = (
        'value',
        '__lock',
    )

    def __init__(self, default_value: _T):
        self.value  = default_value
        self.__lock = Lock()

    def __enter__(self):
        self.__lock.acquire()

    def __exit__(self, *args):
        self.__lock.release()

    def aquire(self):
        self.__lock.acquire()

    def release(self):
        self.__lock.release()

    def bind(self, fn: Callable[typing.Concatenate[typing.Self, _P], _O]) -> Callable[_P, _O]:
        """Return a callable as a method bound to this object. Can be used
        as a decorator."""
        return typing.cast(Callable[_P, _O], types.MethodType(fn, self))

    __call__ = bind




def cfunction(pfunction: 'ctypes._NamedFuncPointer', argtypes: tuple[Any, ...] = (), restype: Any = None):
    """Cast a foreign C function pointer as the decorated function.

    *argtypes* and *restypes* will be assigned to *pfunction*.
    """
    def func_prototype(fn_signature: Callable[_P, _T]) -> Callable[_P, _T]:
        pfunction.argtypes = argtypes
        pfunction.restype  = restype
        functools.update_wrapper(pfunction, fn_signature, updated='')
        return pfunction  # type: ignore
    return func_prototype


# used to cork c-extension memory leaks
@cfunction(ctypes.pythonapi.Py_DecRef, (ctypes.py_object,))
def Py_DECREF(_object: Any) -> None:
    """Reduce the reference count of a Python object by one.

    An object may be deconstructed (garbage collected) when its
    reference count reaches zero.
    """




def timed(fn: Callable[_P, _T]) -> Callable[_P, _T]:
    """Decorator that times the execution of a function."""
    @functools.wraps(fn)
    def wrap(*args: _P.args, **kwds: _P.kwargs) -> _T:
        ts = timer()
        result = fn(*args, **kwds)
        te = timer()
        sig = ', '.join((
            *(repr(v) for v in args),
            *(f"{k}={v!r}" for k,v in kwds.items()),
        ))
        print(f'`{fn.__name__}({sig}) -> ...` took {te-ts} second(s)')
        return result
    timer = time.time
    return wrap



from ._dearpypixl.common import (
    Item,
    ItemAlias,
    ItemUUID,
    ItemCommand,
    Array,
    Literal,
    Sequence,
    Callable,
    Unpack,
    Property,
    Any,
    Color,
    Point,
    overload,
)
from ._dearpypixl.interface import (
    BasicType,
    ContainerType,
    DrawNodeType,
    DrawingType,
    FontType,
    HandlerType,
    NodeEditorType,
    NodeType,
    PlotAxisType,
    PlotType,
    PlottingType,
    RegistryType,
    RootType,
    SupportsCallback,
    SupportsSized,
    SupportsValueArray,
    TableItemType,
    TableType,
    ThemeType,
    WindowType,
    mvAll,
)
from .items import (
    ThemeColor as ThemeColor,
    ThemeStyle as ThemeStyle,
    mvThemeColor as mvThemeColor,
    mvThemeStyle as mvThemeStyle,
)


@overload
def border(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def border(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def border_shadow(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def border_shadow(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def button(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def button(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def button_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def button_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def button_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def button_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def check_mark(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def check_mark(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def child_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def child_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def docking_empty_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def docking_empty_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def docking_preview(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def docking_preview(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def drag_drop_target(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def drag_drop_target(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def frame_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def frame_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def frame_bg_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def frame_bg_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def frame_bg_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def frame_bg_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def header(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def header(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def header_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def header_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def header_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def header_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def menu_bar_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def menu_bar_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def modal_window_dim_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def modal_window_dim_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def nav_highlight(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def nav_highlight(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def nav_windowing_dim_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def nav_windowing_dim_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def nav_windowing_highlight(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def nav_windowing_highlight(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_histogram(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_histogram(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_histogram_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_histogram_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_lines(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_lines(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_lines_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_lines_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def popup_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def popup_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def resize_grip(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def resize_grip(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def resize_grip_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def resize_grip_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def resize_grip_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def resize_grip_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def scrollbar_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def scrollbar_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def scrollbar_grab(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def scrollbar_grab(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def scrollbar_grab_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def scrollbar_grab_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def scrollbar_grab_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def scrollbar_grab_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def separator(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def separator(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def separator_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def separator_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def separator_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def separator_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def slider_grab(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def slider_grab(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def slider_grab_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def slider_grab_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def tab(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def tab(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def tab_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def tab_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def tab_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def tab_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def tab_unfocused(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def tab_unfocused(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def tab_unfocused_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def tab_unfocused_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def table_border_light(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def table_border_light(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def table_border_strong(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def table_border_strong(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def table_header_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def table_header_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def table_row_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def table_row_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def table_row_bg_alt(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def table_row_bg_alt(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def text(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def text(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def text_disabled(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def text_disabled(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def text_selected_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def text_selected_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bg_active(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bg_active(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bg_collapsed(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bg_collapsed(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def window_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def window_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def crosshairs(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def crosshairs(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def error_bar(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def error_bar(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def fill(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def fill(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def inlay_text(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def inlay_text(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def legend_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def legend_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def legend_border(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def legend_border(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def legend_text(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def legend_text(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def line(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def line(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def marker_fill(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def marker_fill(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def marker_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def marker_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_border(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_border(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def plot_frame_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def plot_frame_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def query(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def query(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def selection(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def selection(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_text(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_text(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def x_axis(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def x_axis(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def x_axis_grid(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def x_axis_grid(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis2(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis2(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis3(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis3(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis_grid(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis_grid(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis_grid2(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis_grid2(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def y_axis_grid3(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def y_axis_grid3(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def box_selector(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def box_selector(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def box_selector_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def box_selector_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def grid_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def grid_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def grid_line(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def grid_line(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def grid_line_primary(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def grid_line_primary(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def link(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def link(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def link_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def link_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def link_selected(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def link_selected(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_bg_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_bg_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_canvas(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_canvas(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_canvas_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_canvas_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_link(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_link(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_link_selected(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_link_selected(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_node_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_node_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_node_bg_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_node_bg_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_node_bg_selected(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_node_bg_selected(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_node_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_node_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def mini_map_outline_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def mini_map_outline_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def node_bg(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def node_bg(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def node_bg_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def node_bg_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def node_bg_selected(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def node_bg_selected(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def node_outline(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def node_outline(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def pin(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def pin(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def pin_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def pin_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bar(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bar(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bar_hovered(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bar_hovered(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


@overload
def title_bar_selected(value: Array[int, int, int, int | None], /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...
@overload
def title_bar_selected(r: float, g: float, b: float, a: float = ..., /, *, label: str | None = ..., user_data: Any = ..., use_internal_label: bool = ...) -> mvThemeColor: ...


Border = border
BorderShadow = border_shadow
Button = button
ButtonActive = button_active
ButtonHovered = button_hovered
CheckMark = check_mark
ChildBg = child_bg
DockingEmptyBg = docking_empty_bg
DockingPreview = docking_preview
DragDropTarget = drag_drop_target
FrameBg = frame_bg
FrameBgActive = frame_bg_active
FrameBgHovered = frame_bg_hovered
Header = header
HeaderActive = header_active
HeaderHovered = header_hovered
MenuBarBg = menu_bar_bg
ModalWindowDimBg = modal_window_dim_bg
NavHighlight = nav_highlight
NavWindowingDimBg = nav_windowing_dim_bg
NavWindowingHighlight = nav_windowing_highlight
PlotHistogram = plot_histogram
PlotHistogramHovered = plot_histogram_hovered
PlotLines = plot_lines
PlotLinesHovered = plot_lines_hovered
PopupBg = popup_bg
ResizeGrip = resize_grip
ResizeGripActive = resize_grip_active
ResizeGripHovered = resize_grip_hovered
ScrollbarBg = scrollbar_bg
ScrollbarGrab = scrollbar_grab
ScrollbarGrabActive = scrollbar_grab_active
ScrollbarGrabHovered = scrollbar_grab_hovered
Separator = separator
SeparatorActive = separator_active
SeparatorHovered = separator_hovered
SliderGrab = slider_grab
SliderGrabActive = slider_grab_active
Tab = tab
TabActive = tab_active
TabHovered = tab_hovered
TabUnfocused = tab_unfocused
TabUnfocusedActive = tab_unfocused_active
TableBorderLight = table_border_light
TableBorderStrong = table_border_strong
TableHeaderBg = table_header_bg
TableRowBg = table_row_bg
TableRowBgAlt = table_row_bg_alt
Text = text
TextDisabled = text_disabled
TextSelectedBg = text_selected_bg
TitleBg = title_bg
TitleBgActive = title_bg_active
TitleBgCollapsed = title_bg_collapsed
WindowBg = window_bg
Crosshairs = crosshairs
ErrorBar = error_bar
Fill = fill
InlayText = inlay_text
LegendBg = legend_bg
LegendBorder = legend_border
LegendText = legend_text
Line = line
MarkerFill = marker_fill
MarkerOutline = marker_outline
PlotBg = plot_bg
PlotBorder = plot_border
PlotFrameBg = plot_frame_bg
Query = query
Selection = selection
TitleText = title_text
XAxis = x_axis
XAxisGrid = x_axis_grid
YAxis = y_axis
YAxis2 = y_axis2
YAxis3 = y_axis3
YAxisGrid = y_axis_grid
YAxisGrid2 = y_axis_grid2
YAxisGrid3 = y_axis_grid3
BoxSelector = box_selector
BoxSelectorOutline = box_selector_outline
GridBg = grid_bg
GridLine = grid_line
GridLinePrimary = grid_line_primary
Link = link
LinkHovered = link_hovered
LinkSelected = link_selected
MiniMapBg = mini_map_bg
MiniMapBgHovered = mini_map_bg_hovered
MiniMapCanvas = mini_map_canvas
MiniMapCanvasOutline = mini_map_canvas_outline
MiniMapLink = mini_map_link
MiniMapLinkSelected = mini_map_link_selected
MiniMapNodeBg = mini_map_node_bg
MiniMapNodeBgHovered = mini_map_node_bg_hovered
MiniMapNodeBgSelected = mini_map_node_bg_selected
MiniMapNodeOutline = mini_map_node_outline
MiniMapOutline = mini_map_outline
MiniMapOutlineHovered = mini_map_outline_hovered
NodeBg = node_bg
NodeBgHovered = node_bg_hovered
NodeBgSelected = node_bg_selected
NodeOutline = node_outline
Pin = pin
PinHovered = pin_hovered
TitleBar = title_bar
TitleBarHovered = title_bar_hovered
TitleBarSelected = title_bar_selected


"""Utilities for inspecting Dear PyGui's API."""
import re
import types
import array
import typing
import inspect
import textwrap
import threading
import dataclasses
from inspect import Parameter
from dearpygui import dearpygui
from . import common, constants, tools
from .common import (
    Any,
    Item,
    Callable,
    ItemCommand,
    Sequence,
    Mapping,
    Literal,
)




# [ API Parsers ]

def is_contextmanager(o: Any, /) -> bool:
    """Return True if a function is wrapped by `contextlib.contextmanager`.

    Args:
        * o: The object to inspect.


    >>> from dearpygui import dearpygui
    ...
    >>> is_contextmanager(dearpygui.window)
    True
    >>> is_contextmanager(dearpygui.add_window)
    False
    >>> is_contextmanager(dearpygui.child_window)
    True
    >>> is_contextmanager(dearpygui.add_child_window)
    False
    >>> is_contextmanager(dearpygui.mutex)
    True
    """
    # Works 99% of the time. In theory, it could return false-positives.
    is_ctx_manager = all((
        callable(o),
        o.__globals__["__name__"] == "contextlib",
        o.__module__ != "contextlib",
        isinstance(getattr(o, "__wrapped__", None), types.FunctionType),
    ))
    return is_ctx_manager or False


def is_item_command(o: Any, /) -> bool:
    """Return True if an object is a callable defined in DearPyGui that creates
    and yields/returns an item identifier.

    Only returns True for item commands located in the `dearpygui.dearpygui`,
    not `dearpygui._dearpygui`. Functions in the latter are considered "built-
    in" and have no signature, in which this function returns False.

    Args:
        * o: The object to inspect.


    >>> from dearpygui import dearpygui, _dearpygui
    ...
    >>> is_item_command(dearpygui.window)
    True
    >>> is_item_command(dearpygui.add_window)
    True
    >>> # parsed as a signature-less "built-in"
    >>> is_item_command(_dearpygui.add_window)
    False
    >>> # does not return or accept *tag*
    >>> is_item_command(dearpygui.configure_item)
    False
    >>> def randfunc(*args, tag: int | str = 0) -> int | str: ...
    >>> # not from `dearpygui.dearpygui`
    >>> is_item_command(randfunc)
    False
    """
    # This only works reliably when the function is in `dearpygui`
    # public module because those in the private `_dearpygui` module
    # can't be accurately inspected (no signature).
    try:
        signature = inspect.signature(o)
    except:
        return False  # not callable or is "built-in"

    try:
        if dearpygui.__name__ not in o.__module__:
            return False
    except AttributeError:
        return False

    if signature.return_annotation != int | str:
        return False

    if all(kwd not in signature.parameters for kwd in ("tag", "id")):
        return False
    # weeds out false positives
    elif signature.parameters['tag'].kind != Parameter.KEYWORD_ONLY:
        return False

    # minor optimization for future calls to `signature` (there will be)
    if not hasattr(o, "__signature__"):
        o.__signature__ = signature
    return True








@dataclasses.dataclass(slots=True, frozen=True)
class ItemDefinition:
    name          : str
    enum          : int
    command1      : ItemCommand
    command2      : ItemCommand | None
    init_params   : Mapping[str, Parameter] = dataclasses.field(repr=False)
    rconfig_params: Mapping[str, Parameter] = dataclasses.field(repr=False)
    wconfig_params: Mapping[str, Parameter] = dataclasses.field(repr=False)

    is_container: bool = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        _setattr = object.__setattr__.__get__(self)
        _setattr('is_container', bool(self.command2))


@tools.cache_once
def item_definitions() -> Mapping[str, ItemDefinition]:
    """Parse Dear PyGui's API and return compiled item type definitions.

    The result of this function is only built once and is cached
    for future calls.
    """
    def set_item_types():
        nonlocal name_to_enum
        dearpygui.create_context()
        name_to_enum = dearpygui.get_item_types()
        dearpygui.destroy_context()

    name_to_enum = None
    # This needs to be done in a separate thread so the GPU
    # context isn't created too early or burned for the user.
    _t = threading.Thread(target=set_item_types, daemon=True)
    _t.start()
    _t.join()

    name_to_enum      = common.cast(dict[str, int], name_to_enum)
    name_to_def       = {}
    commands_to_names = {
        dearpygui.add_2d_histogram_series: "mv2dHistogramSeries",
        dearpygui.add_3d_slider          : "mvSlider3D",
        dearpygui.add_hline_series       : "mvHLineSeries",
        dearpygui.add_vline_series       : "mvVLineSeries",
        dearpygui.add_plot_annotation    : "mvAnnotation",
        dearpygui.add_text_point         : "mvLabelSeries",
        dearpygui.draw_image             : 'mvDrawImage',
        dearpygui.draw_text              : 'mvDrawText',
        dearpygui.draw_rectangle         : "mvDrawRect",
        dearpygui.add_subplots           : 'mvSubPlots',
        dearpygui.add_window             : "mvWindowAppItem",
    }
    command_cache = set()
    cpfx_add  = 'add_'
    cpfx_draw = 'draw_'
    for name, obj in dearpygui.__dict__.items():
        if (
            not callable(obj)
            or not is_item_command(obj)
            or "popup" in name
            or obj in command_cache
        ):
            continue

        # find item commands and raw type name
        if is_contextmanager(obj):
            tp_cmd1  = (
                getattr(dearpygui, f"{cpfx_add}{name}", None) or
                getattr(dearpygui, f"{cpfx_draw}{name}")
            )
            tp_cmd2  = obj
            _tp_name = name
        else:
            tp_cmd1  = obj
            _tp_name = name
            # chaining `.removeprefix(...)` could remove more than intended
            if _tp_name.startswith(cpfx_draw):
                _tp_name = _tp_name.removeprefix(cpfx_draw)
            elif _tp_name.startswith(cpfx_add):
                _tp_name = _tp_name.removeprefix(cpfx_add)
            tp_cmd2 = getattr(dearpygui, _tp_name, None)

        command_cache.add(tp_cmd1)
        if tp_cmd2 is not None:
            command_cache.add(tp_cmd2)

        # build final itemtype name
        if tp_cmd1 in commands_to_names:
            tp_name = commands_to_names[tp_cmd1]
            tp_enum = name_to_enum[tp_name]
        else:
            if _tp_name.startswith("colormap"):
                _tp_name = _tp_name.replace("colormap", "color_map", 1)  # capwords
            tp_name = f'mv{_tp_name.title().replace("_", "")}'
            try:
                tp_enum = name_to_enum[tp_name]
            except KeyError:  # `tp_name` is wrong. Try to "fix" it case-by-case.
                # drawing item? (the `.removeprefix` calls from earlier causes this)
                if cpfx_draw in tp_cmd1.__name__ and not tp_name.startswith("mvDraw"):
                    tp_name = f'mvDraw{tp_name.removeprefix("mv")}'
                    tp_enum = name_to_enum[tp_name]
                # item handler? (global handlers shouldn't fail above)
                elif all(s in tp_cmd1.__name__ for s in ("item_", "_handler")):
                    tp_name = tp_name.replace("Item", "", 1)
                    tp_enum = name_to_enum[tp_name]
                # x4 input item?
                elif tp_cmd1.__name__[-1] == "x":
                    tp_name = f"{tp_name[:-1]}Multi"
                    tp_enum = name_to_enum[tp_name]
                else:
                    raise

        # build method parameters
        init_params = upd_param_annotations(
            inspect.signature(tp_cmd1).parameters
        )

        # Ideally, `.configure` should accept what `.configuration` returns
        # as keyword args. But not everything is writable. At the very least,
        # `type(itf)(**.itf.configuration())` MUST be possible so that the
        # interface can be accurately copied/reconstructed.
        # `get_config_parameters` handles a lot of this, but some detail work
        # may need to be done per type.
        rconfig_params = get_config_parameters(init_params)
        if tp_name in ("mvThemeColor", "mvThemeStyle"):
            # `.configure`    : -`target`, -`category`
            # `.configuration`: +`target`, +`category`
            wconfig_params = rconfig_params.copy()
            wconfig_params.pop('category', None)
            rconfig_params['target'] = init_params['target'].replace(
                kind=Parameter.KEYWORD_ONLY
            )
        else:
            wconfig_params = rconfig_params.copy()
        # `pos` is writable as configuration, but is considered a state...
        rconfig_params.pop('pos', '')

        name_to_def[tp_name] = ItemDefinition(
            name=tp_name,
            enum=tp_enum,
            command1=tp_cmd1,
            command2=tp_cmd2,
            init_params=types.MappingProxyType(init_params),
            rconfig_params=types.MappingProxyType(rconfig_params),
            wconfig_params=types.MappingProxyType(wconfig_params),
        )

    return types.MappingProxyType({n:name_to_def[n] for n in sorted(name_to_def)})  # type: ignore




@dataclasses.dataclass(slots=True, frozen=True)
class ElementDefinition:
    name1   : str
    name2   : str
    type    : Literal['mvThemeColor', 'mvThemeStyle']
    constant: str
    target  : int
    category: constants.ThemeCategory



@tools.cache_once
def element_definitions() -> Mapping[str, ElementDefinition]:
    const_pfxs = (
        "mvThemeCol",
        "mvPlotCol",
        "mvNodeCol",
        "mvNodesCol",
        "mvStyleVar",
        "mvPlotStyleVar",
        "mvNodeStyleVar",
        "mvNodesStyleVar",
    )

    CORE = constants.ThemeCategory(dearpygui.mvThemeCat_Core)
    PLOT = constants.ThemeCategory(dearpygui.mvThemeCat_Plots)
    NODE = constants.ThemeCategory(dearpygui.mvThemeCat_Nodes)
    const_to_elem: dict[constants.ThemeCategory, dict[str, ElementDefinition]] = {
        CORE: {},
        PLOT: {},
        NODE: {},
    }
    name_to_elem: dict[constants.ThemeCategory, dict[str, ElementDefinition]] = {
        CORE: {},
        PLOT: {},
        NODE: {},
    }

    RE_CAPS = re.compile(r'([A-Z])')

    for constant, target in dearpygui.__dict__.items():
        pfx, *name1 = constant.split("_", maxsplit=1)
        if pfx not in const_pfxs:
            continue
        name1 = name1[0] \
            .replace("_", "") \
            .replace("Background", "Bg") \
            .replace("bg", "Bg")
        name2 = '_'.join(re.sub(RE_CAPS, r' \1', name1).split()).lower()

        kind = 'mvThemeColor' if pfx.endswith("Col") else 'mvThemeStyle'

        if pfx.startswith('mvPlot'):
            category = PLOT
        elif pfx.startswith('mvNode'):
            category = NODE
        else:
            category = CORE
        name_to_elem[category]

        element = ElementDefinition(
            name1=name1,
            name2=name2,
            type=kind,
            constant=constant,
            target=target,
            category=category,
        )
        const_to_elem[category][constant] = element
        assert name1 not in name_to_elem[category]
        name_to_elem[category][name1] = element

    # Using the names formated above can cause plot and node
    # elements to shadow core elements. Core elements should
    # keep the more simple and intuitive name(s).
    for cat in (PLOT, NODE):
        name_map = name_to_elem[cat]
        for name1, element in name_map.items():
            if name1 in name_to_elem[CORE]:
                name1 = f'{cat.name.title()}{name1}'
                assert name1 not in name_to_elem[CORE]
                assert name1 not in name_map
                object.__setattr__(element, 'name1', name1)
                object.__setattr__(element, 'name2', f'{cat.name.lower()}_{element.name2}')

    for cat in constants.ThemeCategory:
        const_to_elem[cat] = dict(
            sorted(const_to_elem[cat].items(), key=lambda x: x[1].name1)  # type: ignore
        )

    return types.MappingProxyType(
        const_to_elem[CORE] | const_to_elem[PLOT] | const_to_elem[NODE]
    )

@tools.cache_once
def color_definitions() -> Mapping[str, ElementDefinition]:
    return types.MappingProxyType({
        k:v for k,v in element_definitions().items()
        if v.type == "mvThemeColor"
    })

@tools.cache_once
def style_definitions() ->  Mapping[str, ElementDefinition]:
    return types.MappingProxyType({
        k:v for k,v in element_definitions().items()
        if v.type == "mvThemeStyle"
    })








# [ Signature Parsing, Formatting ]

def get_config_parameters(parameters: Mapping[str, Parameter]):
    """Return a mapping of item command parameters containing only
    configuration-related parameters."""
    excluded_names = ("tag", "id", "parent", "default_value",)
    excluded_parts = ("default",)
    return {
        p.name:p for p in parameters.values()
        if p.kind == Parameter.KEYWORD_ONLY
        and p.name not in excluded_names
        and all(npart not in p.name for npart in excluded_parts)
    }


def upd_param_annotations(parameters: Mapping[str, Parameter]) -> dict[str, Parameter]:
    """Return item command parameters with updated annotations."""
    upd_params = {}
    for p in parameters.values():
        if p.kind == Parameter.VAR_KEYWORD:
            continue
        # This specifically targets old Python 3.6 annotations and replaces
        # them with concrete type annotations. Several others are changed
        # for better accuracy.
        name = p.name
        anno = p.annotation

        # case-specific formatting
        if anno == Item:
            anno = 'Item'  # type alias
        elif name == 'label':
            anno = str | None
        elif name == 'pos':
            anno = tuple[int, int] | list[int]
        elif 'callback' in name or name == 'on_close':
            anno = Callable | None
        elif (
            'color' in name
            or (p.default and isinstance(p.default, tuple))
            and len(p.default) == 4
            and anno == typing.List[int] | typing.Tuple[int, ...]
        ):
            anno = tuple[int, int, int, int | None] | list[int]
        elif anno == typing.List[typing.List[float]]:
            anno = Sequence[Sequence[float]]
        # general-case formatting
        else:
            for tp in str, float, int:
                if anno != ((typing.List[tp] | typing.Tuple[tp, ...])):
                    continue
                anno = Sequence[tp]
                break
        upd_params[name] = Parameter(
            name,
            p.kind,
            annotation=anno,
            default=p.default,
        )
    return upd_params



_item_ref  = "A reference to an existing item (as a `int` uuid, `str` alias, or `int` item interface)"
_empty_ref = "None or 0 (default)"

@tools.frozen_namespace
class _DocParseConst:
    RE_ARGS = re.compile(
        r'^\s*Args[;:]$', re.MULTILINE
    )
    RE_PARAM = re.compile(
        r'^\s+(?P<name>\w*)\s?\((?P<type>[^)]+)\):\s*(?P<doc>.*)',
        re.MULTILINE,
    )
    # XXX: This doesn't do the job. Kept as a reference for later (TODO).
    RE_UNION = re.compile(
        r'Union\[\s*(?:[^\s]+]?)(?:(?P<sep>,)\s*[^\s]+)+(?P<end>\])'
    )

    # "large-scope" replacements
    # TODO: regex
    TYPE_REPLACEMENTS_1 = (
        ("Union[int, str]"                      , "int | str"                 ),
        ("Union[List[int], Tuple[int, ...]]"    , f"Sequence[int]"            ),
        ("Union[List[float], Tuple[float, ...]]", f"Sequence[float]"          ),
        ("Union[List[str], Tuple[str, ...]]"    , f"Sequence[str]"            ),
        ("List[List[float]]"                    , f"Sequence[Sequence[float]]"),
        ("List[List[int]]"                      , f"Sequence[Sequence[int]]"  ),
    )
    # "small-scope" replacements
    TYPE_REPLACEMENTS_2 = (
        ("Dict" , "dict" ),
        ("List" , "list" ),
        ("Set"  , "set"  ),
        ("Tuple", "tuple")
    )

    HEAD_SPECIFIC_REPLACEMENTS: dict[str, str] = {}

    PARAM_GENERAL_REPLACEMENTS: dict[str, str] = {
        'before': (
            f"{_item_ref} that is parented by the container that will soon parent this one. If specified, this item will "
            f"be inserted into the parent's child slot preceding the referenced item. If this value is None or 0 (default), "
            f"the item will be added to the end of the slot."
        ),
        'callback': (
            'Called by Dear PyGui when the item is interacted with (usually clicked). Can be None, or any callable with '
            'a `__code__` attribute. Dear PyGui will send up to three positional arguments to the callback; '
            '`sender` (`tag` of the item *callback* is registered to), `app_data` (None, or an updated value associated '
            'with `sender`), and `user_data` (registered *user_data* value of `sender`). However, the callback is not '
            'required to accept all or any of these arguments.'
        ),
        'enabled': (
            "Sets the item's operational state. Setting to False disables the item's functionality. An item's operational "
            "state also affects which elements/components of a theme are applied to it (based off the *enabled_state* value "
            "of each theme component). This setting does not affect the item's visibility state."
        ),
        'filter_key': (
            'The text value of the item that will be used when applying filters.'
        ),
        'height': (
            'Vertical size of the item in pixels. Setting this value may override auto-scaling options for some items; this can be '
            'reversed by setting this back to 0. Note that the value is interpreted as unsigned.'
        ),
        'indent': (
            'Horizontal padding added to the left of the widget. Multiplicative with the `mvStyleVar_IndentSpacing` '
            'theme style element.'
        ),
        'label': (
            'An informal name for the item. Items that are visible in the user interface may also use this as their display '
            'name. The `"##"` delimiter can be used within the label text. If present, only text preceding the first '
            'use of this delimiter will be shown when displaying the label. You can prevent this behavior by escaping the '
            'second hash character with a carriage return (`#\\\\r#`).'
        ),
        'parent': (
            f"{_item_ref} that will become this item's parent. If {_empty_ref}, Dear PyGui will default to "
            f"using whatever item that is on top of the container stack."
        ),
        'show': (
            'Setting to True (default) will render the item and allow it to be seen, while False does the opposite. '
            'For items that cannot be visible in the user interface (handlers, registries, etc), the behavior of '
            '*show* is similar to that of *enabled* where False disables their functionality.'
        ),
        'source': (
            f"{_item_ref}. If specified, getting or setting this item's value will instead get or set the value of the "
            f"referenced item. If the item normally displays its' value, it will instead display the referenced item's "
            f"value."
        ),
        'tag': (
            'An item reference (as a `int` uuid, `str` alias, or `int` item interface) to bind to the interface. If '
            f'{_empty_ref}, or if the integer/string value of the reference is new or unused by another item, '
            "the interface creates a new item of its' type during initialization. If the value is not a new binding "
            "reference, the interface will (try to) operate on the item with an identifier matching that value."
        ),
        'track_offset': (
            'A numeric value between 0 and 1 that affects item tracking while scrolling when *tracked* is True. Common values are'
            ' 0.0 (top), 0.5 (center, default), and 1.0 (bottom).'
        ),
        'tracked': (
            "Enables (True, default) or disables (False) the item's ability to be tracked while scrolling."
        ),
        'use_internal_label': (
            "True (default) will append `'##<item_uuid>'` to the item's *label* value."
        ),
        'user_data': (
            'Can be set as any value or reference. For items with callback-related configuration options, Dear PyGui will send this '
            'as the third positional argument when invoking the callback(s).'
        ),
        'width': (
            f'Horizontal size of the item in pixels. Setting this value may override auto-scaling options for some items; this can be '
            'reversed by setting this back to 0. Note that the value is interpreted as unsigned.'
        ),
    }

    PARAM_SPECIFIC_REPLACEMENTS: dict[str, str] = {}


def upd_command_doc(command: ItemCommand, line_width: int = 92) -> str:
    """Return a formatted docstring parsed from a Dear PyGui item
    command."""
    s = command.__doc__ or ''
    s = s \
        .replace('\t', '    ') \
        .split('    Returns:')[0] \
        .split("    Yields:")[0] \

    key = _DocParseConst.RE_ARGS.search(s)
    if not key:
        return s

    header, body = s.split(key.group(0), maxsplit=1)
    header = textwrap.fill(
        f"{header.strip().rstrip('.')}.",
        line_width,
        subsequent_indent="    "
    )

    # TODO: Build an accurate regex for unions. Harder than it seems -- like
    # reading alphabet soup while drunk.
    for replacement in _DocParseConst.TYPE_REPLACEMENTS_1:
        body = body.replace(*replacement)

    params = array.array('u')
    for param in _DocParseConst.RE_PARAM.finditer(body):
        name, tp, doc = param.groups()
        if '(deprecated)' in doc:
            continue
        if name in _DocParseConst.PARAM_GENERAL_REPLACEMENTS:
            doc = _DocParseConst.PARAM_GENERAL_REPLACEMENTS[name]
        else:
            doc = doc.rstrip().rstrip('.')
            if doc:
                doc = f"{doc}."
        for replacement in _DocParseConst.TYPE_REPLACEMENTS_2:
            tp = tp.replace(*replacement)
        params.fromunicode(textwrap.fill(
            f"* {name} ({tp}): {doc}",
            line_width,
            initial_indent="        ",
            subsequent_indent="        ",
        ))
        params.fromunicode('\n\n')

    return ''.join((
        header,
        "\n\n    Args:\n",
        params.tounicode(),
        f"    Command: `dearpygui.{command.__name__}`"
    ))



import sys
import warnings
import importlib
import importlib.util
import importlib.metadata

# HACK: DPX uses a custom `generate_uuid` function. This patch
# is necessary to ensure the C++ side doesn't auto-increment its
# internal uuid counter (reserved uuids aside). Part 2 is in
# `api.py`.
if 'dearpygui.dearpygui' in sys.modules:
    warnings.warn((
        "Importing `dearpygui` before importing `dearpypixl` may cause "
        "issues. Please consider re-arranging your imports."
    ))

# Get the full path to 'dearpygui.dearpygui'. Unfortunately, this
# requires importing it...
from dearpygui import dearpygui
_fpath = dearpygui.__file__
with open(_fpath, 'r') as file:  # type: ignore
    _patch_src = file.read().replace(
        "tag=tag", "tag=(tag or internal_dpg.generate_uuid())"
    )
del dearpygui  # type: ignore
importlib.invalidate_caches()

dearpygui = importlib.util.module_from_spec(
    importlib.util.spec_from_loader(  # type: ignore
        'dearpygui.dearpygui', None
    )
)
dearpygui.__file__ = _fpath
exec(_patch_src, dearpygui.__dict__)
# The default loader caches submodules in the package namespace.
# It needs to be reset, too.
sys.modules['dearpygui'].dearpygui = sys.modules['dearpygui.dearpygui'] = dearpygui  # type: ignore
